-----------------------------------------------------------------------------
-- EDOM2 -- Nothing spectacular, a lot faster, no compression, cleaner..
----------- Compression handlers, and custom sequence/structure optimizers..
--[Version]
    global constant EDOM_VERSION = 2.03

--[Signs]
                                                --[FollowUP]--[Bits]
    global constant sequence_sign_start = {3,0}        --          -- 4
    global constant sequence_sign_end   = {3,1}        --          -- 4
    global constant boolean_sign_pos    =  1           --          -- 2
    global constant boolean_sign_neg    =  0           --          -- 2
    global constant character_sign_pos  =  2           --          -- 2
    global constant character_sign_neg  = {3,2}        --          -- 4
    global constant integer16_sign_pos  = {3,3,3}      --          -- 6
    global constant integer16_sign_neg  = {3,3,1}      --          -- 6
    global constant integer32_sign_pos  = {3,3,2}      -- 6 bits   -- 6
    global constant integer32_sign_neg  = {3,3,0,1}    -- 6 bits   -- 8
    global constant float32_sign        = {3,3,0,0}    --          -- 8
    global constant float64_sign        = {3,3,0,2}    --          -- 8
    
--[Error-Constants]
    global constant EDO_ERR_FILEIO     = 1      
        -- Unable to open file or unexpected end of file!

    global constant EDO_ERR_VERSION    = 2      
        -- Bad file or wrong version!

    global constant EDO_ERR_INCOMPLETE = 3      
        -- Not all required handlers are loaded!
        
--[Error-Messages]
    global constant EDO_ERRORS = { 
            "Unable to open file or unexpected end of file!",
            "Bad file or wrong EDOM version!",
            "Not all custom add-ons used are available now!"
                              }
            

--[Compression]
    sequence cp_names       -- All the names of the compression handlers
    sequence cp_iDin        -- The input iD's
    sequence cp_iDout       -- The output iD's
    sequence cp_iDopen      -- The open iD's
    sequence cp_iDclose     -- The close iD's 
    sequence cp_iDwrite     -- The uncompressed output iD's 
    sequence cp_iDread      -- The uncompressed input iD's
    integer out, in         -- Current iD's

--[Register_cp]
    global function register_cp (sequence n, integer din, integer dout, 
 integer dopen, integer dclose, integer dwrite, integer dread, integer dseek)
    integer dummy_i
                                         
        -- Removing -[eol]- values.
        while find ('\n', n) do
            n = n[1..find('\n',n)-1] & n[find('\n',n)+1..length(n)]
        end while
                                     
        -- Registering all information about this CP
        cp_names = append(cp_names, n)
        cp_iDin = append(cp_iDin, (din=-1)*cp_iDin[1])+((din!=-1)*din)
        cp_iDout = append(cp_iDout, (dout=-1)*cp_iDout[1])+((dout!=-1)*dout)
        cp_iDopen = append(cp_iDopen, (dopen=-1)*cp_iDopen[1])+((dopen!=-1)*dopen)
        cp_iDclose = append(cp_iDclose, (dclose=-1)*cp_iDclose[1])+((dclose!=-1)*dclose)
        cp_iDwrite = append(cp_iDwrite, (dwrite=-1)*cp_iDwrite[1])+((dwrite!=-1)*dwrite)
        cp_iDread = append(cp_iDread, (dread=-1)*cp_iDread[1])+((dread!=-1)*dread)
        -- Shortforms: append id, unless id = -1, then append NONE-id (1st)
        
        dummy_i = dseek     -- Not yet used !
        
        -- Return the cp-id
        return length(cp_names)
        
    end function      
    
--[CPHandler: None]
    sequence stack  stack = {}
    object dummy
    integer start   start = 0
    
    global function none_in (integer fh, integer c)
    
        if start then
            stack = stack[start..length(stack)]
            start = 0
        end if
        
        if c = -1 then
            if length(stack) then
                start = length(stack)+1
                return stack
            else                     
                return gets(fh)
            end if
        elsif c = -2 then              
            if not length(stack) then
                dummy = gets(fh)
                if sequence(dummy) then
                    stack = dummy
                else
                    return {}
                end if
            end if
            while stack[length(stack)]+1 do
                stack = stack & gets(fh)
            end while
            start = length(stack) + 1
            return stack
        end if
        
        if c then
            while c > length(stack) do
                stack = stack & gets (fh)
            end while   
            start = c + 1
            return stack[1..c]    
        else  
            if length(stack) then
                start = 2
                return stack[1]
            else
                return getc(fh)
            end if
        end if
        
    end function
    
    global procedure none_out (integer fh, object x)
        puts(fh, x)
    end procedure
    
    global function none_open (sequence fname, sequence opt)
        return open(fname, opt)
    end function               
    
    global procedure none_close (integer fh)
        close (fh)
    end procedure                     
    
    global procedure none_write (integer fh, object x)
        puts(fh, x)
    end procedure
    
    global function none_read (integer fh)
        return gets (fh)
    end function
    
    global function none_seek (integer fh, integer p)
        return machine_func(19, {fh, p})
    end function
    
--[Structure: Default]
    global function default_struct_shrink (sequence s, sequence s2)
    -- S2 is not used cuz it could help us at all.
    sequence ret     
    integer i
    
        -- Initialising i (counter for ret)
        i = 0                              
        
        -- Adding to s if needed
        if and_bits(3,length(s)) then
            s = s & repeat(0, 4 - and_bits(3,length(s)))
        end if
        
        -- Preparing return sequence
        ret = repeat (0, floor(length(s)/4))
        
        -- Combing all 2-bits into bytes
        for index = 1 to length(s)-1 by 4 do
            i = i + 1        
            ret[i] = s[index] + s[index+1] * 4 
                   + s[index+2] * 16 + s[index+3] * 64
        end for   
        
        return ret
        
        -- This is never executed, but it avoids an unneeded warning
            s2 = s2     
        -- Robert! With and without should be LOCAL (and global as an option)
    end function                    

    global function default_struct_expand (sequence s)
    sequence ret         
    integer i, val
        
        -- Preparing return sequence
        ret = repeat (0, length(s)*4)
        
        -- Initialising i (counter for s)
        i = 0
        
        for index = 1 to length (ret) by 4 do
            i = i + 1
            val = s[i]
            ret[index] = and_bits(3,val)
            val = floor(val/4)
            ret[index+1] = and_bits(3,val)
            val = floor(val/4)
            ret[index+2] = and_bits(3,val)
            ret[index+3] = floor(val/4)
        end for
        
        return ret
    end function                                              
    
--[Management: Structures]

    sequence structure_name, old_structure
    integer struct_id_shrink, struct_id_expand,   -- Handlers
            structure_bytenize
    
    old_structure = {}                            -- Old structure lifo stack

    -- You need to load the structure before it will be used
    global procedure load_structure (sequence name, integer in, integer out, 
                                                                integer bt)
        
        old_structure = append(old_structure, {name, in, out, not bt})
        
        if not find('\n', name) then
            name = name & '\n'
        else
            name = name[1..find('\n',name)]
        end if
        structure_name = name

        struct_id_shrink = out
        struct_id_expand = in
        structure_bytenize = not bt
        
    end procedure
    
    -- When you unload an structure the previous structure is used
    global procedure unload_structure ()
        -- Storing none structure iD
        if length(old_structure) < 1 then
            struct_id_shrink = routine_id("default_struct_shrink")
            struct_id_expand = routine_id("default_struct_expand")  
            structure_bytenize = 1
            structure_name = "DEFAULT\n"
        else                    
            structure_bytenize = old_structure[1][4]
            struct_id_shrink = old_structure[1][3]
            struct_id_expand = old_structure[1][2]
            structure_name = old_structure[1][1]
            old_structure = old_structure[2..length(old_structure)]
        end if                                  
    end procedure
    
--[Management: SO]
    sequence so_names         so_names = {}
    sequence so_shrink        so_shrink = {}
    sequence so_expand        so_expand = {}

    global procedure load_so (sequence name, integer shrink, integer expand)
        if not find('\n', name) then
            name = name & '\n'
        else
            name = name[1..find('\n',name)]
        end if
        if not find(name, so_names) then
            so_names = append (so_names, name)
            so_shrink = append(so_shrink, shrink)
            so_expand = append(so_expand, expand)
        else
            so_shrink[find(name, so_names)] = shrink
            so_expand[find(name, so_names)] = expand
        end if
    end procedure                                   
    
    global procedure unload_so (sequence name)
    integer ifind
        ifind = find(name, so_names)
        if ifind then
            so_names = so_names[1..ifind-1] & so_names[ifind+1..length(so_names)]
            so_shrink= so_shrink[1..ifind-1] & so_shrink[ifind+1..length(so_shrink)]
            so_expand = so_expand[1..ifind-1] & so_expand[ifind+1..length(so_expand)]
        end if      
    end procedure    
    
--[Locals]
    integer fhandle         -- File Handler currently in use
    
    -- An edo is an combination of DATA and a STRUCTURE seq. (both 1D seq)
    sequence structure, data
    
--[EDO_Save]
    constant log256 = log(256)  -- Needed for bytes to use calculation
                                -- Because it is fixed, this speeds up..!
        
    -- A recursive procedure that splits the seq up into data & structure
    procedure store_sequence (sequence s)
    sequence float32_s
    integer val
        for index = 1 to length(s) do                                        
            if integer(s[index]) then
                val = s[index] 
                if val < 0 then
                    if val > -257 then
                        structure = structure & character_sign_neg
                        data = append(data, val+1)
                    elsif val > -65793 then
                        structure = structure & integer16_sign_neg
                        data = append(append(data,val),floor(val/256)+1)
                    else -- Replace with 32-bit int neg!
                        val = -val
                        structure = structure & integer32_sign_neg
                        data =append(append(append(append(data, val), floor(val/256)), floor(val/65536)),floor(val/ 16777216))
                    end if
                elsif val < 2 and structure_bytenize then
                    structure = append(structure,val)
                elsif val < 258 then              
                    structure = structure & character_sign_pos
                    data = append(data,val)
                elsif val < 65793 then
                    structure = structure & integer16_sign_pos
                    data = append(append(data,val-1),floor(val/256)-1)
                else -- Replace with 32-bit int pos!
                    structure = structure & integer32_sign_pos
                    data =append(append(append(append(data, val), floor(val/256)), floor(val/65536)),floor(val/ 16777216))
                end if
            elsif sequence(s[index]) then
                structure = structure & sequence_sign_start
                store_sequence (s[index])                  
                structure = structure & sequence_sign_end
            else                                     
                float32_s = machine_func (48,s[index]) 
                if machine_func (49, float32_s) then
                    structure = structure & float32_sign
                    data = data & float32_s
                else
                    structure = structure & float64_sign
                    data = data & machine_func(46,s[index])
                end if
            end if
        end for     
    end procedure                 

    -- The actual saving routine that does the trick
    global function edo_save (sequence fname, sequence s, integer cp)
    integer nob, temp_i -- number of bytes needed, temporary integer
    
        -- Open file
        fhandle = call_func (cp_iDopen[cp], {fname, "wb"})
        if fhandle = -1 then return EDO_ERR_FILEIO end if
        
        -- Check for empty sequence (BUG FIX!!)
        if not length(s) then
            call_proc (cp_iDwrite[cp], {fhandle, "EDOM2EMPTY\n"})    
            call_proc (cp_iDclose[cp], {fhandle})
            return 0        -- No error occured!
        end if                                                  
        
        -- Write out recognizable tag
        call_proc (cp_iDwrite[cp], {fhandle, "EDOM2\n"})
        
        -- Write out compression method
        call_proc (cp_iDwrite[cp], {fhandle, cp_names[cp] & '\n'})
        
        -- Setting handlers
        out = cp_iDout[cp]
        
        -- Writing out structure optimizer
        call_proc (out, {fhandle, structure_name})
        
        -- Write out the type of optimizers and execute them
        for index = 1 to length(so_names) do
            call_proc(out, {fhandle, so_names[index]})
            s = call_func(so_shrink[index],{s})
        end for
        call_proc(out, {fhandle, "EOSO\n"})
        
        -- Procces the sequence into structure, data
        structure = {}  data = {}
        store_sequence (s)

        -- Execute structure optimizer
        structure = call_func (struct_id_shrink, {structure,s})
        
        -- Releasing memory
        s = {}
        
        -- Saving the structure length to disk
        nob = floor(log(length(structure))/log256)+1
        call_proc (out, {fhandle, nob & length(structure)})
        
        if nob > 1 then
            temp_i = floor (length(structure) / #100)
            call_proc (out, fhandle & temp_i)
            for index = 3 to nob do
                temp_i = floor (temp_i / #100)
                call_proc (out, fhandle & temp_i)
            end for
        end if
            
        -- Save the structure to disk (write out the structure)
        call_proc (out, {fhandle, structure})
        
        -- Releasing memory
        structure = {}
        
        -- Save the data to disk (write out the data)
        call_proc (out, {fhandle, data})
        
        -- Releasing memory
        data = {}
        
        -- Stopping/logging off compression
        call_proc (cp_iDclose[cp], {fhandle})
        
        -- Saved ? Yes, No error occured (false)
        return 0
        
    end function              

--[EDO_Load]

    -- The gets argument for the IN-handlers:
    -- The {fn, -1} & {fn, 1} & {fn, -2} combi's (gets, getc, get-to-eof)
    sequence gss, gcs, gas

    -- Indexes
    integer data_index, structure_index, endc

    -- Recursive function that interpreters the data & structure seq-s
    function load_sequence ()     
    sequence oret, ret     
    integer counter
        oret = {}  
        ret = repeat (0, 255)
        while 1 do       
            counter = 0                                              
            while counter < 255 and structure_index < endc do
            counter = counter + 1
            if structure[structure_index] < 3 then -- Charcater/Boolean
                 if structure[structure_index] < 2 then
                    ret[counter] = structure[structure_index]
                    structure_index = structure_index + 1
                else  
                    ret[counter] = data[data_index]
                    data_index = data_index + 1
                    structure_index = structure_index + 1
                end if
            else -- other
                if structure[structure_index+1] < 2 then
                    if structure[structure_index+1] then
                        structure_index = structure_index + 2
                        if length(oret) then     
                            if length(ret) = counter-1 then
                                return oret & ret
                            else
                                return oret & ret[1..counter-1]
                            end if
                        else      
                            if length(ret) = counter-1 then
                                return ret
                            else
                                return ret[1..counter-1]
                            end if
                        end if
                    else                                     
                        structure_index = structure_index + 2
                        if length(oret) then
                            oret = oret & ret[1..counter-1]
                        else
                            oret = ret[1..counter-1]
                        end if
                        oret = append(oret, load_sequence())
                        counter = 0
                    end if
                elsif structure[structure_index+1] > 2 then
                    if structure[structure_index+2] then
                        if structure[structure_index+2] < 2  then -- 1
                            ret[counter] = data[data_index]-256 + (data[data_index+1]-256)*256
                            structure_index = structure_index + 3
                            data_index = data_index + 2
                        elsif structure[structure_index+2] > 2 then -- 3
                            ret[counter] = data[data_index]+1+(data[data_index+1]+1)*256
                            structure_index = structure_index + 3
                            data_index = data_index + 2
                        else
                            ret[counter] = data[data_index]
                                         + data[data_index+1] * 256
                                         + data[data_index+2] * 65536
                                         + data[data_index+3] *  16777216
                            structure_index = structure_index + 3
                            data_index = data_index + 4
                        end if
                    elsif structure[structure_index+3] < 1 then -- 0
                        ret[counter] = machine_func(49, data[data_index..data_index+3])
                        data_index = data_index + 4
                        structure_index = structure_index + 4
                    elsif structure[structure_index+3] > 1 then -- 2
                        ret[counter] = machine_func(47, data[data_index..data_index+7])
                        data_index = data_index + 8
                        structure_index = structure_index + 4
                    else    -- Replace with 32-bit integer neg! 
                            ret[counter] = -(data[data_index]
                                            + data[data_index+1] * 256
                                            + data[data_index+2] * 65536
                                            + data[data_index+3] *  16777216)
                            structure_index = structure_index + 4
                            data_index = data_index + 4
                    end if
                else
                    ret[counter] = data[data_index]-1
                    data_index = data_index + 1
                    structure_index = structure_index + 2
                end if
            end if             
        end while
            
        if structure_index >= endc then
            if length(oret) then     
                if length(ret) = counter then
                    return oret & ret
                else
                    return oret & ret[1..counter]
                end if
            else      
                if length(ret) = counter then
                    return ret
                else
                    return ret[1..counter]
                end if
            end if
        end if
        if length(ret) = counter then
            oret = oret & ret
        else
            oret = oret & ret[1..counter]
        end if
        end while
        
    end function

    -- The actual routine that does the trick..
    global function edo_load (sequence fname, integer cp)
    sequence readin
    integer nob, temp_i, temp_i2

        -- Open file
        fhandle = call_func(cp_iDopen[cp], {fname, "rb"})
        if fhandle = -1 then return EDO_ERR_FILEIO end if
        gss = {fhandle,-1} gcs = {fhandle, 0} gas = {fhandle, -2}
        
        -- Check recognizable tag (BUG FIX!!)
        readin = call_func(cp_iDread[cp], {fhandle})
        if compare (readin, "EDOM2\n") then
            if compare (readin, "EDOM2EMPTY\n") then
                return EDO_ERR_VERSION
            else         
                call_proc(cp_iDclose[cp], {fhandle})
                return {}
            end if
        end if
        
        -- Read in compression type
        readin = call_func(cp_iDread[cp], {fhandle})
        if readin[length(readin)] != '\n' then return EDO_ERR_FILEIO end if
        cp = find(readin[1..length(readin)-1], cp_names)
        if cp then
            in = cp_iDin[cp]
        else
            return EDO_ERR_INCOMPLETE
        end if
        
        -- Reading in structure type
        readin = call_func(in, gss)
        if compare(structure_name, readin) then
            return EDO_ERR_INCOMPLETE 
        end if
        
        -- Read in SO's...
        readin = {call_func(in, gss)}
        while compare(readin[length(readin)], "EOSO\n") do
            readin = append(readin, call_func(in, gss))
            if not find ('\n', readin[length(readin)]) then
                return EDO_ERR_FILEIO
            end if
        end while
        
        -- Getting number-of-bytes of structure length filed
        nob = call_func (in, gcs)
        
        -- Checking EOF
        if nob = -1 then return EDO_ERR_FILEIO end if
        
        -- Reading in structure length field
        temp_i = call_func (in, gcs)
        temp_i2 = 256
        for counts = 2 to nob - 1 do
            temp_i = temp_i + call_func(in, gcs) * temp_i2
            temp_i2 = temp_i2 * 256
        end for
        if nob > 1 then temp_i = temp_i + call_func(in, gcs) * temp_i2 end if

        -- Reading in structure
        structure = call_func (in, {fhandle,temp_i})
        
        -- Checking for EOF
        if find (-1, structure) then return EDO_ERR_FILEIO end if
        
        -- Expanding structure sequence
        structure = call_func (struct_id_expand, {structure})      
        
        -- Reading in the data
        data = call_func(in, gas)
        
        -- Closing filehandle
        call_proc (cp_iDclose[cp],{fhandle})
        
        -- Setting the indexes to the beginning
        data_index = 1                                                
        structure_index = 1
        if length(structure) then
            endc = length(structure) + 1 - (1 * (structure[length(structure)]=-1)) - 1
        else 
            return EDO_ERR_FILEIO
        end if             
        if length(data) then
            if data[length(data)] = -1 then
                data = data[1..length(data)-1]
            end if
        end if
        
        -- Generating return sequence

        return load_sequence ()
                    
    end function

--[Initializing]
    cp_names = {"NONE"}
    cp_iDout = {routine_id("none_out")}
    cp_iDin = {routine_id("none_in")}
    cp_iDopen = {routine_id("none_open")}
    cp_iDclose = {routine_id("none_close")}  
    cp_iDwrite = {routine_id("none_write")}
    cp_iDread = {routine_id("none_read")}
    
    -- Setting the NONE-compression as a global constant
    global constant CP_NONE = 1                       

    -- Initialising structure optimization to DEFAULT
    unload_structure ()
    
-----------------------------<<.. Ralf Nieuwenhuijsen, nieuwen@xs4all.nl ..>>


