--=-=-=-=-=-=-=-=         - Ralf Nieuwenhuijsen -
-- EDOMpak.EX   =              - E.D.O.M -
--=-=-=-=-=-=-=-=                 - 2 - 
-- An example for EDOM2 which bundles a set of files to a .EDO
-- The structure handler is custom.. (a fixed structure)
-- A sequence handler is not used (the structure and data do not relate!)

--<<.. Welcome message ..>>

    puts(1, "\n EDOMPAK 1.0 by Ralf Nieuwenhuijsen (c)    \n")
    puts(1,   " ---                                       \n")
    puts(1,   " Use freely for any non-commercial purpose!\n")
    puts(1,   " Use at own risk!                          \n")
    puts(1,   " ---                                       \n")
    puts(1,   " Send any bug reports to nieuwen@xs4all.nl \n")
    puts(1,   " Also for fanmail, comments and mailbombs! \n")
    puts(1,   " ---                                       \n")

--<<.. Includes ..>>
    include edom.e      -- Duh ? What you expected!
    without warning             -- Keeps complaining about an argument that has to be there
                                -- due to calling convention of structure handlers.

--<<.. Display version ..>>
    printf(1, " Using EDOM version: %2.2f                 \n", {EDOM_VERSION})
    puts(1,   " ---                                       \n") 
    
--<<.. Constants ..>>
    constant comline = command_line ()
    
    if length(comline) < 4 then
        puts(1,   " USAGE: edompak e|c filename.edo file1 file2 ..      \n")
        puts(1,   " ---                                                 \n")
        puts(1,   " Use the command e to extracts the chosen files.     \n")
        puts(1,   " Use the command c to compress the chosen files.     \n")
        puts(1,   " ---                                                 \n") 
        abort(1)
    end if                                                   
    
    function make_len12 (sequence s)          
    
        for index = 1 to length(s) do 
            if length(s[index]) > 12 then
                s[index] = s[index][1..12]
            elsif length(s[index]) < 12 then
                s[index] = s[index] & repeat (' ', 12-length(s[index]))
            end if
        end for
        
        return s
    end function
    
    -- Splitting up arguments
    constant files = make_len12( comline[5..length(comline)] )
    constant edo_file = comline [4]                           
    constant command = comline [3][1] -- First char of first argument
    
--<<.. Making the data sequence ..>>

    sequence data,    -- All data from all the files
             newline  
             
    procedure load_files ()         -- The layout of this routine may look like shit in DOS
    integer  fhandle
    
        data = repeat ({}, length(files)+1)
        data[1] = files
                for index = 1 to length(files) do
                            fhandle = open(files[index], "rb")
                            if fhandle = -1 then
                                    puts (1,   " Unable to open: " & files[index] & '\n')
                                data[index+1] = {}
                            else
                                        puts (1,   " Succesfully opened: " & files[index] & '\n')
                                         newline = gets(fhandle)
                                 while newline[length(newline)] != -1 do
                                                newline = newline & gets(fhandle)
                                            end while            
                                            close(fhandle)                                 
                                    data[index+1] = newline
                        end if
                end for
        end procedure
    
    procedure write_files ()
    integer i, fhandle
        
        for index = 1 to length(files) do
            i = find (files[index], data[1])
            fhandle = open (files[index], "wb")
            if fhandle = -1 then
                puts (1,   " Unable to write: " & files[index] & '\n')
            else
                puts (fhandle, data[index+1])
                close (fhandle)              
                puts (1,   " Succesfully written: " & files[index] & '\n')
            end if                          
        end for
        
    end procedure

--<<.. The structure handler ..>>

    include machine.e
    
    function struct_fileset_shrink (sequence st, sequence sd)
    -- The st gets the structure sequence (see edomhelp.ex)
    -- The sd gets the original data you want to save or load
    sequence ret                  
        
        ret = {}
        for index = 2 to length(sd) do
            -- Writing out length of file..
            ret = ret & int_to_bytes (length(sd[index]))
        end for
        
        return ret
    end function                          
    
    constant log2 = log(2)
    
    function repeat_seq (sequence o, integer i)
    sequence ret, a      
                
        ret = {}
        while i do
            a = o         
            for index = 1 to log (i) / log2 do
                a = a & a
            end for            
            i = i - length(a) / length(o)
            ret = ret & a
            if i = 1 then return ret & o end if
        end while
        
        return ret
    end function  

    function struct_fileset_expand (sequence s)
    sequence ret, lens

        -- Variable step loop through 'S'  
        lens = {}
        for index = 1 to length(s) by 4 do
            lens = lens & bytes_to_int(s[index..index+3])
        end for  
        
        -- Making file-list structure
        ret = sequence_sign_start    
            & repeat_seq (
                        sequence_sign_start
                            & repeat(character_sign_pos, 12)
                        & sequence_sign_end
                        ,
                        length(lens)
                         )
            & sequence_sign_end 
            
        -- Making file-data structure
        for index2 = 1 to length(lens) do
            ret = ret 
                & sequence_sign_start
                    & repeat(character_sign_pos, lens[index2]) 
                & sequence_sign_end
        end for                                      
        
        return ret  -- It's that simple 8-)
    end function      
    
    load_structure ("PAK1", 
                            routine_id("struct_fileset_expand"),  
                            routine_id("struct_fileset_shrink"),  1)
                            
    
--<<.. Let's (de)compress ..>>

    object error
    
    if command = 'c' or command = 'C' then
        load_files ()
        puts (1, " ---\n")
        error = edo_save (edo_file, data, CP_NONE)
        if error then 
            puts (1, EDO_ERRORS[error])
        else 
            puts (1, " Succesfully saved: " & edo_file & '\n')
        end if    
    else
        trace (1)
        error = edo_load (edo_file, CP_NONE)
        if sequence(error) then
            data = error             
            puts (1, " Succesfully loaded: " & edo_file & '\n')
            puts (1, " ---\n")
            write_files ()
        else  
            puts (1, ' ' & EDO_ERRORS[error] & '\n')
        end if
    end if  
    
--<<.. End message ..>>

    puts (1,   " --- \n")

--<<.. End of demonstration program ..>>---------------------------------->>



